/*
 * Copyright (c) 1993-2012 CNRS/LAAS
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS.
 */

/**
 ** gyroServerLib.c
 **
 ** Functions to open, read and decode gyro data on a serial line :
 **
 **      int gyroOpen(char *device)
 **      int gyroRead(int fd, char *buf, uint32_t *date)
 **      int gyroDecode(GyroData *gyro)
 **
 ** Author: Anthony Mallet (original gyro.c) and Sara Fleury
 ** Date:
 **
 **/

#include <err.h>
#include <errno.h>
#include <math.h>
#include <pthread.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "serial.h"
#include "gyroPrivate.h"
#include "KVH1000.h"

static void *gyroKVH1000ReadTask(void *);
static int gyroReadData(int fd, char *, uint32_t *);
static int gyroDecode(struct GYRO_KVH1000_DATA *);
static int gyroClose(struct GYRO_DATA *);

/* ----------------------------------------------------------------------
 *
 * gyroInit -     Init connection to serial port
 *
 *
 * Returns :            FALSE if failed or TRUE.
 */

struct GYRO_DATA *
gyroInitKVH1000(const char *port, double latitude, double woffset)
{
	struct GYRO_KVH1000_DATA *gyroData;
	int rv;

	gyroData = (struct GYRO_KVH1000_DATA *)malloc(sizeof(struct GYRO_KVH1000_DATA));
	if (gyroData == NULL)
		return NULL;
	pthread_mutex_init(&gyroData->gd.lock, NULL);
	pthread_mutex_lock(&gyroData->gd.lock);
	if ((gyroData->fd = openSerial(port, GYRO_KVH1000_SERIAL_SPEED)) == -1) {
		warn("openSerial %s", port);
		return NULL;
	}
	gyroData->gd.wOffset = woffset;
	gyroData->stop = 0;
	gyroData->gd.earthSpeed = gyroEarthSpeed(latitude);
	gyroData->gd.close = gyroClose;
	rv = pthread_create(&gyroData->taskId, NULL,
	    gyroKVH1000ReadTask, gyroData);
	pthread_mutex_unlock(&gyroData->gd.lock);
	if (rv != 0) {
		pthread_mutex_destroy(&gyroData->gd.lock);
		free(gyroData);
		return NULL;
	}
	return (struct GYRO_DATA *)gyroData;
}


/*
 * gyroClose ------------------------------------------------------------
 */
static int
gyroClose(struct GYRO_DATA *p)
{
	struct GYRO_KVH1000_DATA *gyro = (struct GYRO_KVH1000_DATA *)p;

	gyro->stop = 1;
	if (close(gyro->fd) == -1)
		return -1;
	pthread_mutex_destroy(&gyro->gd.lock);
	free(gyro);
	return 0;
}

/*
 * gyroRead -------------------------------------------------------------
 *
 * Read a message
 */

static int
gyroReadData(int fd, char *buf, uint32_t *msec)
{
   int i;

   /* read 8 bytes */
   for(i=0; i<8; i++) {
	   if (read(fd, &buf[i], 1) <= 0) {
		   fprintf(stderr, "gyroRead: error read %s\n",
		       strerror(errno));
		   return -1;
	   }
   }

  sync:
   /* continue read until a message header is found (8th bit of first byte
    * must be set) */
   while(!(buf[0] & 0x80)) {
      printf("gyroRead: skipping one byte\n");

      memcpy(buf, buf+1, 7);
      if (read(fd, buf+7, 1) <= 0) return -1;
   }

   /* check message is complete (8th bit of the 7 last bytes must be
    * unset) */
   for(i=1; i<8; i++) if (buf[i] & 0x80) {
      printf("gyroRead: incomplete message\n");

      memcpy(buf, buf+1, 7);
      goto sync;
   }

   *msec = getTime();

   return 0;
}


/*
 * gyroDecode -----------------------------------------------------------
 *
 * Decode a raw message
 */

int
gyroDecode(struct GYRO_KVH1000_DATA *gyro)
{
   double w;
   int d;
   char *buf = gyro->buf;
   unsigned char c;

   /* check checksum */
   c  = (buf[0]<<1) + ((buf[1]&0x7F)>>6);
   c += (buf[1]<<2) + ((buf[2]&0x7F)>>5);
   c += (buf[2]<<3) + ((buf[3]&0x7F)>>4);
   c += (buf[3]<<4) + ((buf[4]&0x7F)>>3);
   c += (buf[4]<<5) + ((buf[5]&0x7F)>>2);
   c += (buf[5]<<6) + ((buf[6]&0x7F)>>1);
   c += (buf[6]<<7) + (buf[7]&0x7F);
   if (c) {
      fputs("gyro: bad checksum\n", stderr);
      return -1;
   }

   /* check "built-in-test" bit */
   if (!(buf[3] & 0x20)) {
      fputs("gyro: built-in-test failed\n", stderr);
      return -1;
   }

   /* serial number */
   if (!gyro->snO) {
      if (buf[3] & 0x40) {
	 if (!gyro->snI) {
	    /* first bit */
	    gyro->snBit = 1;
	    gyro->snI = 1;
	    gyro->sn = 0;
	 } else {
	    /* last bit */
	    gyro->snBit = 0;
	    gyro->snO = 1;
	 }
      }

      if (buf[2] & 0x01) gyro->sn |= gyro->snBit;
      gyro->snBit <<= 1;

      if (gyro->snO) {
	printf("gyro serial number : %d\n", gyro->sn);
      }
   }

   /* temperature */
   if (buf[4] & 0x4)
      gyro->tlsb = ((buf[4]<<5)&0x7F) + ((buf[5]&0x7F)>>2);
   else {
      gyro->tmsb = ((buf[4]<<5)&0x7F) + ((buf[5]&0x7F)>>2);

      d = ((((short)gyro->tmsb)<<7) + gyro->tlsb) &0xFFF;
      if (d & 0x800) d += 0xFFFFF000;

      gyro->temp = d * GYRO_KVH1000_TEMP_RESOLUTION;
   }

   /* rate */

#if VERBOSE>1
   printf("%02x %02x %02x", buf[5], buf[6], buf[7]);
#endif

   d = (((short)(buf[5]&0x3))<<14)+(((short)buf[6]&0x7F)<<7)+(buf[7]&0x7F);
   d &= 0xFFFF;
   if (d & 0x8000) d += 0xFFFF0000;
   w = d * GYRO_KVH1000_RATE_RESOLUTION - gyro->gd.earthSpeed;
#if VERBOSE>1
   printf(", rate=%04x %f rad/s", d, w);
#endif


   gyroIntegrate((struct GYRO_DATA *)gyro, w);
#if VERBOSE>0
   printf(", %5.4f deg/s, dt=%u ms, th=%5.4f deg\n",
	  RAD_TO_DEG(gyro->w), gyro->gd.period, RAD_TO_DEG(gyro->gd.theta));
#endif

   return 0;
}

/*
 * gyroReadTask ---------------------------------------------------------
 *
 * This function reads and decode gyro data permanently.
 *
 * Thus it must be started within an independant thread.
 *
 */

static void*
gyroKVH1000ReadTask(void *param)
{
	struct GYRO_KVH1000_DATA *gyro = param;
	uint32_t date;

	fputs("---- gyro: started\n", stderr);

	while(!gyro->stop) {
		if (gyroReadData(gyro->fd, gyro->buf, &date) < 0) break;

		pthread_mutex_lock(&gyro->gd.lock);
		gyro->gd.date = date;
		if (gyroDecode(gyro) < 0)
			fputs("gyro: lost message\n", stderr);
		pthread_mutex_unlock(&gyro->gd.lock);
	}

	/* Stop */
	if (close(gyro->fd) == -1) {
		printf ("gyro: cannot close fd %d errno = %d \n", gyro->fd, errno);
	}

	pthread_mutex_destroy(&gyro->gd.lock);
	free(gyro);

	fputs("gyro: stopped\n", stderr);
	return NULL;
}
